/* (*+ newsstats NEWSSTATS='-pop -in NSData.NSData'*) */
/* NewsStats
   To make the now infamous Posting Statistics list ;-)
   RjY, (c) anARCHy, 2000-2001

   0.00 (21:2/2/2000) BASIC version
   0.10 (23:2/2/2000) C version
        (24:2/2/2000) took it to the unix labs and finished it off.
   0.20 (16:2/3/2000) the "generalised" version, handles name and mail address changes.
   0.21 (23:2/3/2000) at most one "From:" line per post
   0.30 (18:1/1/2001) Last Week's Position In Brackets
        (21:1/1/2001) finalised new format for table
   0.40 (21:3/1/2001) Internal changes (linked list, not array) "Minimal width"
   0.41 (20:2/2/2001) Poster list self-organises as it is built
   0.50 (26:2/2/2001) Does popularity lists. Removed matching options.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXINP 1023

int Popularity;

typedef struct article_s {
  struct article_s *prev, *next;
  char *id, *from;
  int replies; /* No. of posts that are replies to this one */
} article;

article acap = {&acap, &acap, NULL, NULL, 0};
int total_articles = 0;

typedef struct poster_s {
  struct poster_s *prev, *next;
  int pos, prevpos, count, prevcount, replies;
  char *user, *host, *name;
} poster;

poster pcap = {&pcap, &pcap, 0, 0, 0, 0, 0, NULL, NULL, NULL};
int total_posters = 0;

typedef struct {
  unsigned char *user, *host, *name;
  int userl, hostl, namel;
} fromline;

typedef struct {
  poster *poster;
  int count;
} matches;

void nomemory(void)
{
  fprintf(stderr, "newsstats: Not enough memory\n");
}

int checkparameter(int count, char **args, const char *parm)
{
  int i;

  for (i=1; i<count; i++)
    if (!strcmp(args[i], parm))
      return i;
  return 0;
}

void readoptions(int argcount, char **args)
{
  int p;

  if (p=checkparameter(argcount, args, "-pop")) Popularity=1; else Popularity=0;
}

/* if first n chars match and one of the strings has
length n>0, return true(1), else return false(0) */
int comparefromstrs(const char *s1, const char *s2, const int n)
{
  return (n > 0 && (s1[n]=='\0' || s2[n]=='\0') && !strncmp(s1,s2,n));
}

void parsefromline(char *string0, fromline *from)
{
  char *string=string0, *at=NULL, *word0, *word1;
  char *host=NULL, *user=NULL, *name=NULL;
  int hostl=0, userl=0, namel=0;

  if (!strncmp(string0,"From: ",6)) string+=6;

  word0 = string;
  while (1)
  {
    for (word1 = word0; *word1 > ' '; word1++)
      if (at == NULL && *word1 == '@') at = word1;

    if (word0 <= at && at < word1)
    {
      user = word0; userl = (at-word0);
      host = at+1;  hostl = (word1-host);
    }
    else
    {
      if (name == NULL) name = word0;
      namel = word1-name;
    }

    if (*word1 < ' ')
      break;
    else
    {
      word0 = word1;
      while (*word0 == ' ') word0++;
    }
  }

  if (userl > 0 && hostl > 0 && user[0] == '<' && host[hostl-1] == '>')
    { user++; userl--; hostl--; }
  if (namel > 0 && ((name[0] == '('  && name[namel-1] == ')')
   || (name[0] == '\"' && name[namel-1] == '\"')))
    { name++; namel--; namel--; }

  from->user = user; from->host = host; from->name = name;
  from->userl = userl; from->hostl = hostl; from->namel = namel;
}

int addnewposter(fromline *from, poster *newptemplate)
{
  poster *p, *newp;
  int length;
  char *user,*host,*name;

  if (!(newp = malloc(sizeof(poster)))
    || !(user = malloc(from->userl+1))
    || !(host = malloc(from->hostl+1))
    || !(name = malloc(from->namel+1)))
  {
    nomemory(); return 0;
  }

  *newp = *newptemplate;

  strncpy(newp->user=user, from->user, length=from->userl); newp->user[length]='\0';
  strncpy(newp->host=host, from->host, length=from->hostl); newp->host[length]='\0';
  strncpy(newp->name=name, from->name, length=from->namel); newp->name[length]='\0';

  p = pcap.next;
  newp->prev = &pcap; newp->next = p;
  pcap.next = newp; p->prev = newp;

  return 1;
}

matches compareposters(fromline *from)
{
  poster *p;
  matches matches;

  matches.count = 0; matches.poster = &pcap;
  for (p = pcap.next; p != &pcap; p = p->next)
  {
    if (comparefromstrs(p->user, from->user, from->userl)
      || comparefromstrs(p->name, from->name, from->namel))
    {
      matches.count++; matches.poster = p;
    }
  }
  return matches;
}

int scanprevious(FILE *prf)
{
  char prid[MAXINP+1];
  poster newp;
  fromline from;
  matches matches;

  memset(&newp, 0, sizeof(poster));
  while (1)
  {
    fscanf(prf,"%d %*s %d %*s %[^\n]", &newp.prevpos, &newp.prevcount, prid);
    if (feof(prf)) return 1;
    parsefromline(prid, &from);
    if (!addnewposter(&from, &newp)) return 0;
  }
}

void handleprevious(int argcount, char **args)
{
  int p;
  FILE *file;

  if (!Popularity)
  {
    p = checkparameter(argcount, args, "-prev");
    if (p && (p < argcount-1))
    {
      if (!(file = fopen(args[p+1], "r")))
      {
        fprintf(stderr, "newsstats: cannot open prev file %s\n",args[p+1]);
        exit(1);
      }
      if (!scanprevious(file)) exit(1);
      fclose(file);
    }
  }
}

void updateparent(char *refs)
{
  char *parentid = refs;
  int i;
  article *ar;

  for (i = 0; refs[i] != '\0'; i++)
    if (refs[i] == ' ') parentid = refs+i+1;
  for (ar = acap.prev; ar != &acap; ar = ar->prev)
    if (!strcmp(parentid, ar->id)) ar->replies++;
}

int buildarticlelist(FILE *inf)
{
  char string[MAXINP+1], *id = NULL, *from = NULL;
  article *ar = NULL, *tmp;
  int foundrefs = 0;

  fgets(string,MAXINP,inf);
  do
  {
    if (!id && !strncmp(string,"Message-ID: ",12))
    {
      if (!(id = malloc(strlen(string)+1)))
        { nomemory(); return 0; }
      else
        strcpy(id,string+12);
    }
    else if (!from && !strncmp(string,"From: ",6))
    {
      if (!(from = malloc(strlen(string)+1)))
        { nomemory(); return 0; }
      else
        strcpy(from,string+6);
    }
    else if (!foundrefs && !strncmp(string,"References: ",12))
    {
      foundrefs = 1; updateparent(string+12);
    }
    else if (!strcmp(string,".\n"))
    {
      if (id != NULL && from != NULL)
      {
        /* Add article to front i.e. construct list "backwards" */
        ar = malloc(sizeof(article));
        ar->id = id; ar->from = from; ar->replies = 0;
        tmp = acap.next;
        ar->prev = &acap; ar->next = tmp;
        acap.next = ar; tmp->prev = ar;
      }
      else if (id != NULL)
        free(id);
      else if (from != NULL)
        free(from);
      ar = NULL; id = NULL; from = NULL; foundrefs = 0;
    }
    fgets(string,MAXINP,inf);
  } while (!feof(inf));
  return 1;
}

int buildposterlist(void)
{
  article *ar;
  fromline from;
  poster *p, *tmp, newp;
  matches matches;

  memset(&newp, 0, sizeof(poster));
  for (ar = acap.next; ar != &acap; ar = ar->next)
  {
    parsefromline(ar->from, &from);
    if (!from.userl && !from.hostl && !from.namel)
      continue; /* ignore badly formatted From: lines */
    total_articles++;
    matches = compareposters(&from);
    if (matches.count == 0)
    {
      newp.count = 1; newp.replies = ar->replies; total_posters++;
      if (!addnewposter(&from, &newp)) return 0;
    }
    else if (matches.count == 1)
    {
      p = matches.poster;
      p->count++; p->replies += ar->replies;
      if ((tmp = pcap.next) != p)
      {
        p->prev->next = p->next; p->next->prev = p->prev;
        p->next = tmp; p->prev = &pcap;
        pcap.next = p; tmp->prev = p;
      }
    }
  }
  return 1;
}

void handleinput(int argcount, char **args)
{
  int p;
  FILE *file;

  p=checkparameter(argcount, args, "-in");
  while (p && (p < argcount-1) && (args[p+1][0] != '-') )
  {
    if (!(file = fopen(args[p+1], "r")))
    {
      fprintf(stderr, "newsstats: cannot open data file %s\n",args[p+1]);
      exit(1);
    }
    if (!(buildarticlelist(file))) exit(1);
    fclose(file); p++;
  }
  if (!buildposterlist()) exit(1);
}

#if 0
void debug_printarticlelist()
{
  article *ar;
  int c = 0;

  for (ar = acap.next; ar != &acap; ar = ar->next)
  {
    printf("Id: %sFrom: %sReplies: %d\n",ar->id,ar->from,ar->replies);
    c++;
  }
  printf("\n%d articles (or %d)\n\n",c,total_articles);
}

void debug_printposterlist()
{
  poster *p;
  int c = 0;

  for (p = pcap.next; p != &pcap; p = p->next)
  {
    printf("Poster: %s <%s@%s>\nReplies: %d\nCount: %d\n",
      p->name,p->user,p->host,p->replies,p->count);
    c++;
  }
  printf("\n%d posters (or %d)\n\n",c,total_posters);
}
#endif

void sortresults(void)
{
  poster *p,*q,*tmp;
  int swap,cp,cq,prevcp,prevcq,pos;
  double mean, popp, popq;

  mean = (Popularity || total_posters)
    ? ((double)total_articles / (double)total_posters) : 0.0 ;

  for (pos = 1, p = pcap.next; p != &pcap; pos++, p = p->next)
  {
    for (q = p; q != &pcap; q = q->next)
    {
      if (!Popularity)
      {
        cp = p->count; cq = q->count; prevcp = p->prevcount; prevcq = q->prevcount;
        swap = (cp<cq
          || (cp == cq && prevcp < prevcq)
          || (cp == cq && prevcp == prevcq && strcmp(p->name, q->name)>0));
      }
      else
      {
        cp = p->count; cq = q->count;
        popp = (2.0 * p->replies) / (mean + cp);
        popq = (2.0 * q->replies) / (mean + cq);
        swap = (popp < popq);
      }
      if (swap)
      {
        tmp = q->prev;
        q->prev->next = q->next; q->next->prev = q->prev;
        q->next = p; q->prev = p->prev;
        p->prev->next = q; p->prev = q;
        p = q; q = tmp;
      }
    }
    if (pos == 1 || p->count != p->prev->count)
      p->pos = pos;
    else
      p->pos = p->prev->pos;
  }
}

int digits(int n)
{
  int i,c;

  if (n<10) return 1;
  if (n<100) return 2;
  for (i=2, c=100; c<=n; i++, c=10*c);
  return i;
}

void printresults(void)
{
  static const char NEW[] = "(-)";
  poster *p;
  int maxpos, maxprevpos = 1, maxcount, maxprevcount = 1;
  char format[20], prevpos_b[8], prevcount_b[8], *prevpos, *prevcount;

  if (!Popularity)
  {
    maxcount = pcap.next->count;
    for (p = pcap.next; p->count > 0 && p != &pcap; p = p->next)
    {
      maxprevpos = p->prevpos > maxprevpos ? p->prevpos : maxprevpos;
      maxprevcount = p->prevcount > maxprevcount ? p->prevcount : maxprevcount;
    }
    maxpos = p->prev->pos;

    sprintf(format, "%%%dd %%%ds %%%dd %%%ds ",
      digits(maxpos), digits(maxprevpos)+2,
      digits(maxcount), digits(maxprevcount)+2);
  }
  else
  {
    maxpos = 0; maxcount = 0;
    for (p = pcap.next; p != &pcap; p = p->next)
    {
      if (p->replies > maxpos) maxpos = p->replies;
      if (p->count > maxcount) maxcount = p->count;
    }
    sprintf(format, "%%%ds ", digits(maxpos)+digits(maxcount)+1);
  }

  for (p = pcap.next; p->count > 0 && p != &pcap; p = p->next)
  {
    if (!Popularity)
    {
      if (p->prevpos)
        sprintf(prevpos = prevpos_b,"(%d)",p->prevpos);
      else
        (const) prevpos = NEW; /* dodgy - but surpresses compiler warnings */
      if (p->prevcount)
        sprintf(prevcount = prevcount_b,"(%d)",p->prevcount);
      else
        (const) prevcount = NEW;
      printf(format, p->pos, prevpos, p->count, prevcount);
    }
    else
    {
      sprintf(prevpos_b, "%d:%d", p->replies, p->count);
      printf(format, prevpos_b);
    }

    if (*p->user && *p->host && *p->name)
      printf("%s <%s@%s>\n", p->name, p->user, p->host);
    else if (*p->user && *p->host && !(*p->name))
      printf("%s@%s\n", p->user, p->host);
    else
      printf("%s\n", p->name);
  }
}

int main(int argcount, char **args)
{
  readoptions(argcount, args);

  handleprevious(argcount, args);
  handleinput(argcount, args);

/*   debug_printarticlelist(); */
/*   debug_printposterlist(); */

  sortresults();
  printresults();

  return 0;
}
